@@ -12,6 +12,7 @@
 #include "evolve_kepler.h"
 #include "evolve_ok.h"
 #include "evolve_bs.h"
+#include "evolve_error_control.h"
 
 #ifdef EVOLVE_OPENCL
 #include "evolve_cl.h"
@@ -19,9 +20,13 @@
 
 int verbosity=0;
 
+struct sys debugsys;
+
 FLOAT eps2;
 FLOAT dt_param;
-struct sys zerosys ={ 0, NULL,NULL};
+struct sys zerosys ={0,0,NULL,NULL};
+int accel_zero_mass=1;
+int opencl_device_type=0;
 
 struct diagnostics global_diag;
 struct diagnostics *diag;
@@ -34,12 +39,14 @@ static void report(struct sys s,DOUBLE etime, int inttype);
 
 void move_system(struct sys s, DOUBLE dpos[3],DOUBLE dvel[3],int dir)
 {
+  struct particle *ipart;
   for(UINT p=0;p<s.n;p++)
   {
+    ipart=GETPART(s, p);
     for(int i=0;i<3;i++)
     {
-        COMPSUMP(s.part[p].pos[i],s.part[p].pos_e[i],dir*dpos[i])
-        COMPSUMV(s.part[p].vel[i],s.part[p].vel_e[i],dir*dvel[i])
+        COMPSUMP(ipart->pos[i],ipart->pos_e[i],dir*dpos[i])
+        COMPSUMV(ipart->vel[i],ipart->vel_e[i],dir*dvel[i])
     }
   }  
 }
@@ -47,14 +54,16 @@ void move_system(struct sys s, DOUBLE dpos[3],DOUBLE dvel[3],int dir)
 void system_center_of_mass(struct sys s, DOUBLE *cmpos, DOUBLE *cmvel)
 {
   DOUBLE mass=0.,pos[3]={0.,0.,0.},vel[3]={0.,0.,0.};
+  struct particle *ipart;
   for(UINT p=0;p<s.n;p++)
   {
+    ipart=GETPART(s, p);
     for(int i=0;i<3;i++)
     {
-      pos[i]+=(DOUBLE) s.part[p].mass*s.part[p].pos[i];
-      vel[i]+=(DOUBLE) s.part[p].mass*s.part[p].vel[i];
+      pos[i]+=(DOUBLE) ipart->mass*ipart->pos[i];
+      vel[i]+=(DOUBLE) ipart->mass*ipart->vel[i];
     }
-    mass+=(DOUBLE) s.part[p].mass;
+    mass+=(DOUBLE) ipart->mass;
   }
   for(int i=0;i<3;i++)
   {
@@ -63,23 +72,32 @@ void system_center_of_mass(struct sys s, DOUBLE *cmpos, DOUBLE *cmvel)
   }
 }
 
-FLOAT system_kinetic_energy(struct sys s)
+DOUBLE system_kinetic_energy(struct sys s)
 {
  UINT i;
  DOUBLE e=0.;
- for(i=0;i<s.n;i++) e+=0.5*s.part[i].mass*(
-                           s.part[i].vel[0]*s.part[i].vel[0]+
-                           s.part[i].vel[1]*s.part[i].vel[1]+
-                           s.part[i].vel[2]*s.part[i].vel[2]);
- return (FLOAT) e;
+ struct particle *ipart;
+ for(i=0;i<s.n;i++)
+ {
+   ipart=GETPART(s, i);
+   e+=0.5*ipart->mass*( ipart->vel[0]*ipart->vel[0]+
+                        ipart->vel[1]*ipart->vel[1]+
+                        ipart->vel[2]*ipart->vel[2] );
+ }
+ return e;
 }
         
-FLOAT system_potential_energy(struct sys s)
+DOUBLE system_potential_energy(struct sys s)
 {
  UINT i;
  DOUBLE e=0.;
- for(i=0;i<s.n;i++) e+=s.part[i].mass*s.part[i].pot;
- return (FLOAT) e/2;
+ struct particle *ipart;
+ for(i=0;i<s.n;i++)
+ {
+   ipart=GETPART(s, i);
+   e+=ipart->mass*ipart->pot;
+ }
+ return e/2;
 }
 
 void init_code()
@@ -100,18 +118,20 @@ void stop_code()
 
 void init_evolve(struct sys s,int inttype)
 {
-  UINT i;
-  for(i=0;i<s.n;i++)
+  struct particle *ipart;
+  for(UINT i=0;i<s.n;i++)
   {
-    s.part[i].postime=0.;
-    s.part[i].pot=0.;
+    ipart=GETPART(s,i);
+    ipart->postime=0.;
+    ipart->pot=0.;
 #ifdef COMPENSATED_SUMMP
-    s.part[i].pos_e[0]=0.;s.part[i].pos_e[1]=0.;s.part[i].pos_e[2]=0.;
+    ipart->pos_e[0]=0.;ipart->pos_e[1]=0.;ipart->pos_e[2]=0.;
 #endif
 #ifdef COMPENSATED_SUMMV
-    s.part[i].vel_e[0]=0.;s.part[i].vel_e[1]=0.;s.part[i].vel_e[2]=0.;
+    ipart->vel_e[0]=0.;ipart->vel_e[1]=0.;ipart->vel_e[2]=0.;
 #endif
   }
+  if(accel_zero_mass) split_zeromass(&s); // because of potential calc
   potential(s,s);
   
   evolve_ok_stop();
@@ -171,20 +191,21 @@ void sum_diagnostics(struct diagnostics* total,struct diagnostics* diag)
   total->cl_count+=diag->cl_count;
 #endif
 #ifdef _OPENMP
-  printf("%d: %d %li %li %li\n",omp_get_thread_num(),tasksum,diag->taskdrift,
-     diag->taskkick,taskcountsum);
+  if(verbosity>0) printf("task %d: %d %li %li %li\n",omp_get_thread_num(),tasksum,diag->taskdrift, diag->taskkick,taskcountsum);
 #endif
 }
 
 
 void do_evolve(struct sys s, double dt, int inttype)
 {
-  UINT p;
   int i,clevel;
+  struct particle *ipart;
   if(dt==0) return;
-  for(p=0;p<s.n;p++) s.part[p].postime=0.;
+  for(UINT p=0;p<s.n;p++) GETPART(s,p)->postime=0.;
   clevel=0;
+  if(accel_zero_mass) split_zeromass(&s);
   zero_diagnostics(diag);
+  debugsys=s;
   switch (inttype)
   {
     case CONSTANT:
@@ -204,22 +225,22 @@ void do_evolve(struct sys s, double dt, int inttype)
       evolve_constant10(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt);
       break;
     case SHARED2:
-      evolve_shared2(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
+      evolve_shared2(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
       break;
     case SHARED4:
-      evolve_shared4(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
+      evolve_shared4(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
       break;
     case SHARED6:
-      evolve_shared6(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
+      evolve_shared6(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
       break;
     case SHARED8:
-      evolve_shared8(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
+      evolve_shared8(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
       break;
     case SHARED10:
-      evolve_shared10(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
+      evolve_shared10(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
       break;
     case SHAREDBS:
-      evolve_bs_adaptive(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
+      evolve_bs_adaptive(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
       break;
     case BS_CC_KEPLER:
       evolve_bs(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt);
@@ -256,17 +277,26 @@ void do_evolve(struct sys s, double dt, int inttype)
     case CCC_KEPLER:
     case CCC_BS:
     case CCC_BSA:
+    case CC_SHARED10:
+    case CCC_SHARED10:
 #ifdef _OPENMP
 #pragma omp parallel shared(global_diag,s,dt,clevel) copyin(dt_param) 
       {
         diag=(struct diagnostics *) malloc(sizeof( struct diagnostics));
         zero_diagnostics(diag);
+#pragma omp master      
+	      if(verbosity>0) printf("Total Threads # %d\n", omp_get_num_threads()); 
 #pragma omp single
 #endif
+#ifdef CC2_SPLIT_SHORTCUTS
+        evolve_cc2_shortcut(clevel,s,(DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, inttype, 1, -1.);
+#else
         evolve_cc2(clevel,s,(DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, inttype, 1);
+#endif        
 #ifdef _OPENMP
 #pragma omp critical
         sum_diagnostics(&global_diag,diag);
+        free(diag);
       }
       diag=&global_diag;
 #endif        
@@ -275,8 +305,7 @@ void do_evolve(struct sys s, double dt, int inttype)
       evolve_ok2(clevel,s, zeroforces, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
       break;
     case KEPLER:
-      if(s.n==2) evolve_kepler(clevel,s,(DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt);
-      else evolve_kepler_test(clevel,s,(DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt);
+      evolve_kepler(clevel,s,(DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt);
       break;
     case FOURTH_M4:
       evolve_sf_4m4(clevel,s,(DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt,1);
@@ -299,24 +328,28 @@ void do_evolve(struct sys s, double dt, int inttype)
     case SHARED10_COLLISIONS:
       evolve_shared10_collision_detection(s, (DOUBLE) dt);
       break;
+    case ERROR_CONTROL:
+      evolve_error_control(clevel,s, (DOUBLE) 0.,(DOUBLE) dt,(DOUBLE) dt, -1.);
+      break;
     default:  
-      ENDRUN("unknown integrator\n");
+      ENDRUN("unknown integrator %d\n", inttype);
       break;
   } 
-  for(p=0;p<s.n;p++) s.part[p].pot=0;
+  for(UINT p=0;p<s.n;p++) GETPART(s,p)->pot=0;
   potential(s,s);
   if(verbosity>0) report(s,(DOUBLE) dt, inttype);
 }
 
 void drift(int clevel,struct sys s, DOUBLE etime, DOUBLE dt)
 {
-  UINT i;
-  for(i=0;i<s.n;i++)
+  struct particle *ipart;
+  for(UINT i=0;i<s.n;i++)
   {
-    COMPSUMP(s.part[i].pos[0],s.part[i].pos_e[0],dt*s.part[i].vel[0])
-    COMPSUMP(s.part[i].pos[1],s.part[i].pos_e[1],dt*s.part[i].vel[1])
-    COMPSUMP(s.part[i].pos[2],s.part[i].pos_e[2],dt*s.part[i].vel[2])
-    s.part[i].postime=etime;
+    ipart=GETPART(s,i);
+    COMPSUMP(ipart->pos[0],ipart->pos_e[0],dt*ipart->vel[0])
+    COMPSUMP(ipart->pos[1],ipart->pos_e[1],dt*ipart->vel[1])
+    COMPSUMP(ipart->pos[2],ipart->pos_e[2],dt*ipart->vel[2])
+    ipart->postime=etime;
   }
   diag->dstep[clevel]++;
   diag->dcount[clevel]+=s.n;
@@ -325,47 +358,49 @@ void drift(int clevel,struct sys s, DOUBLE etime, DOUBLE dt)
 
 static void kick_cpu(struct sys s1, struct sys s2, DOUBLE dt)
 {
-  UINT i,j;
   FLOAT dx[3],dr3,dr2,dr,acci;
   FLOAT acc[3];
+  struct particle *ipart, *jpart;
 
-#pragma omp parallel for if((ULONG) s1.n*s2.n>MPWORKLIMIT && !omp_in_parallel()) default(none) \
- private(i,j,dx,dr3,dr2,dr,acc,acci) \
+#pragma omp parallel for if((ULONG) s1.n*(s2.n-s2.nzero)>MPWORKLIMIT && !omp_in_parallel()) default(none) \
+ private(dx,dr3,dr2,dr,acc,acci,ipart,jpart) \
  shared(dt,s1,s2,eps2)
-  for(i=0;i<s1.n;i++)
+  for(UINT i=0;i<s1.n;i++)
   {
+    ipart=GETPART(s1,i);
     acc[0]=0.;
     acc[1]=0.;
     acc[2]=0.;
-    for(j=0;j<s2.n;j++)
+    for(UINT j=0;j<s2.n-s2.nzero;j++)
     {
-      if(s2.part[j].mass==0) continue;
-//      if(s1.part+i==s2.part+j) continue; 
-      dx[0]=s1.part[i].pos[0]-s2.part[j].pos[0];
-      dx[1]=s1.part[i].pos[1]-s2.part[j].pos[1];
-      dx[2]=s1.part[i].pos[2]-s2.part[j].pos[2];
+      jpart=GETPART(s2,j);
+//~ if(jpart->mass==0) continue;
+//      if(ipart==jpart) continue; 
+      dx[0]=ipart->pos[0]-jpart->pos[0];
+      dx[1]=ipart->pos[1]-jpart->pos[1];
+      dx[2]=ipart->pos[2]-jpart->pos[2];
       dr2=dx[0]*dx[0]+dx[1]*dx[1]+dx[2]*dx[2]+eps2;
       if(dr2>0) 
       {
         dr=sqrt(dr2);
         dr3=dr*dr2;
-        acci=s2.part[j].mass/dr3;
+        acci=jpart->mass/dr3;
 
         acc[0]-=dx[0]*acci;
         acc[1]-=dx[1]*acci;
         acc[2]-=dx[2]*acci;  
       }
     }
-    COMPSUMV(s1.part[i].vel[0],s1.part[i].vel_e[0],dt*acc[0]);
-    COMPSUMV(s1.part[i].vel[1],s1.part[i].vel_e[1],dt*acc[1]);
-    COMPSUMV(s1.part[i].vel[2],s1.part[i].vel_e[2],dt*acc[2]);
+    COMPSUMV(ipart->vel[0],ipart->vel_e[0],dt*acc[0]);
+    COMPSUMV(ipart->vel[1],ipart->vel_e[1],dt*acc[1]);
+    COMPSUMV(ipart->vel[2],ipart->vel_e[2],dt*acc[2]);
   }
 }
 
 void kick(int clevel,struct sys s1, struct sys s2, DOUBLE dt)
 {
 #ifdef EVOLVE_OPENCL
-  if((ULONG) s1.n*s2.n>CLWORKLIMIT) 
+  if((ULONG) s1.n*(s2.n-s2.nzero)>CLWORKLIMIT) 
   {
 #pragma omp critical
     kick_cl(s1,s2,dt);
@@ -387,37 +422,39 @@ void kick(int clevel,struct sys s1, struct sys s2, DOUBLE dt)
 
 static void potential_cpu(struct sys s1,struct sys s2)
 {
-  UINT i,j;
   FLOAT dx[3],dr2,dr;
   FLOAT pot;
+  struct particle *ipart, *jpart;
 
-#pragma omp parallel for if((ULONG) s1.n*s2.n>MPWORKLIMIT && !omp_in_parallel()) default(none) \
- private(i,j,dx,dr2,dr,pot) \
+#pragma omp parallel for if((ULONG) s1.n*(s2.n-s2.nzero)>MPWORKLIMIT && !omp_in_parallel()) default(none) \
+ private(dx,dr2,dr,pot, ipart,jpart) \
  shared(s1,s2,eps2)
-  for(i=0;i<s1.n;i++)
+  for(UINT i=0;i<s1.n;i++)
   {
     pot=0;
-    for(j=0;j<s2.n;j++)
+    ipart=GETPART(s1,i);
+    for(UINT j=0;j<s2.n-s2.nzero;j++)
     {
-      if(s1.part+i==s2.part+j) continue; 
-      dx[0]=s1.part[i].pos[0]-s2.part[j].pos[0];
-      dx[1]=s1.part[i].pos[1]-s2.part[j].pos[1];
-      dx[2]=s1.part[i].pos[2]-s2.part[j].pos[2];
+      jpart=GETPART(s2,j);
+      if(ipart==jpart) continue; 
+      dx[0]=ipart->pos[0]-jpart->pos[0];
+      dx[1]=ipart->pos[1]-jpart->pos[1];
+      dx[2]=ipart->pos[2]-jpart->pos[2];
       dr2=dx[0]*dx[0]+dx[1]*dx[1]+dx[2]*dx[2]+eps2;
       if(dr2>0) 
       {
         dr=sqrt(dr2);
-        pot-=s2.part[j].mass/dr;
+        pot-=jpart->mass/dr;
       }
     }
-    s1.part[i].pot+=pot;
+    ipart->pot+=pot;
   }
 }
 
 void potential(struct sys s1, struct sys s2)
 {
 #ifdef EVOLVE_OPENCL
-  if((ULONG) s1.n*s2.n>CLWORKLIMIT) 
+  if((ULONG) s1.n*(s2.n-s2.nzero)>CLWORKLIMIT) 
   {
 #pragma omp critical
     potential_cl(s1,s2);
@@ -477,28 +514,31 @@ inline FLOAT timestep_ij(struct particle *i, struct particle *j,int dir) {
 
 static void timestep_cpu(struct sys s1, struct sys s2,int dir)
 {
-  UINT i,j;
+  UINT i,j, jmax;
   FLOAT timestep,tau;
-#pragma omp parallel for if((ULONG) s1.n*s2.n>MPWORKLIMIT && !omp_in_parallel()) default(none) \
- private(i,j,tau,timestep) copyin(dt_param) \
+  struct particle *ipart;
+#pragma omp parallel for if((ULONG) (s1.n*s2.n-s1.nzero*s2.nzero)>MPWORKLIMIT && !omp_in_parallel()) default(none) \
+ private(i,j,tau,timestep, jmax, ipart) copyin(dt_param) \
  shared(s1,s2,stdout,dir)
   for(i=0;i<s1.n;i++)
   {  
     timestep=HUGE_VAL;
-    for(j=0;j<s2.n;j++)
+    ipart=GETPART(s1,i);
+    jmax=s2.n;if(i>=s1.n-s1.nzero) jmax=s2.n-s2.nzero;
+    for(j=0;j<jmax;j++)
     {
-      tau=timestep_ij(s1.part+i,s2.part+j,dir);
+      tau=timestep_ij(ipart,GETPART(s2,j),dir);
       if(tau < timestep) timestep=tau;
     }
-//    if(timestep<s1.part[i].timestep) 
-    s1.part[i].timestep=timestep;
+//    if(timestep<ipart->timestep) 
+    ipart->timestep=timestep;
   }
 }
 
 void timestep(int clevel,struct sys s1, struct sys s2,int dir)
 {
 #ifdef EVOLVE_OPENCL
-  if((ULONG) s1.n*s2.n>CLWORKLIMIT) 
+  if((ULONG) (s1.n*s2.n-s1.nzero*s2.nzero)>CLWORKLIMIT) 
   {
 #pragma omp critical
     timestep_cl(s1,s2,dir);
@@ -518,6 +558,7 @@ static void report(struct sys s,DOUBLE etime, int inttype)
   int maxlevel=0,i;
   long int ktot=0,dtot=0, kstot=0,dstot=0,ttot=0,tstot=0;
   UINT n,p,err=0;
+  struct particle *ipart;
   n=s.n;
   printf("** report **\n");
   printf("interaction counts:\n");
@@ -541,14 +582,12 @@ static void report(struct sys s,DOUBLE etime, int inttype)
   printf("steps: %18li, equiv: %18li, maxlevel: %i\n", 
     diag->deepsteps,((long) 1)<<maxlevel,maxlevel); 
 
-  for(p=0;p<s.n;p++)
-  {
-    if(s.part[p].postime != (DOUBLE) etime) err++;
-  }
+  for(p=0;p<s.n;p++) if(GETPART(s,p)->postime != (DOUBLE) etime) err++;
   printf("postime errors: %u \n",err);
   printf("target time, actual time: %12.8g %12.8g %12.8g\n", 
            (double) etime,(double) diag->simtime,(double) ((DOUBLE) etime-diag->simtime));
-  printf("time track, ratio: %12.8g %12.8g\n", (double) diag->timetrack,(double) (diag->timetrack/diag->simtime));
+  printf("time track, ratio: %12.8g %12.8g\n", (double) diag->timetrack,
+       (double) (diag->simtime!=0? (diag->timetrack/diag->simtime) :1));
 
 #ifdef EVOLVE_OPENCL
   printf("cpu step,count: %12li,%18li\n",diag->cpu_step,diag->cpu_count);
@@ -566,7 +605,8 @@ static void report(struct sys s,DOUBLE etime, int inttype)
     }
     printf(" total, total j, mean j: %18li %18li %f\n",totalbs,totalj,totalj/(1.*totalbs));
   }
-  if(inttype==KEPLER || inttype==CC_KEPLER || inttype==CCC_KEPLER)
+  if(inttype==KEPLER || inttype==CC_KEPLER || inttype==CCC_KEPLER || inttype==CCC_BS || 
+     inttype==CC_BS || inttype==CCC_BSA || inttype==CC_BSA || inttype==CC_SHARED10 || inttype==CC_SHARED10)
   {
     unsigned long totalcefail=0,totalcecount=0;
     printf("kepler solver counts:\n");
@@ -588,49 +628,80 @@ static void report(struct sys s,DOUBLE etime, int inttype)
       printf("%d: %18li %18li\n",i,diag->ntasks[i],diag->taskcount[i]);
       totaltasks+=diag->ntasks[i];
     } 
-   printf("openmp tasks: %d\n",totaltasks);
-
+    printf("openmp tasks: %d\n",totaltasks);
   }
-
 #endif  
   fflush(stdout);
 }
 
+void join_array(UINT n1, struct particle *p1,
+                UINT n2, struct particle *p2,
+                UINT *n, struct particle **p)
+{
+  if(n1==0 && n2==0)
+  {
+    *n=0;*p=NULL;
+  }
+  if(n1!=0 && n2==0)
+  {
+    *n=n1;*p=p1;
+  }
+  if(n1==0 && n2!=0)
+  {
+    *n=n2;*p=p2;
+  }
+  if(n1!=0 && n2!=0)
+  {
+    *n=n1+n2;
+    if(p1+n1==p2)
+    {
+      *p=p1;
+    }
+    else
+    {
+      if(p2+n2==p1)
+      {
+        *p=p2;
+      } else ENDRUN("join_array error");
+    }
+  }
+}                  
+
 struct sys join(struct sys s1,struct sys s2)
 {
   struct sys s=zerosys;
-  if(s1.n == 0) return s2;
-  if(s2.n == 0) return s1;  
-  s.n=s1.n+s2.n;
-  if(s1.part+s1.n == s2.part)
-  {
-    s.part=s1.part;
-    s.last=s2.last;
-  } else
-  {
-    if(s2.part+s2.n == s1.part)
-    {
-      s.part=s2.part;
-      s.last=s1.last;
-    } else
-      ENDRUN("join error 1");
-  }   
-  if(s.last-s.part + 1 != s.n) ENDRUN("join error 2");
+  if(s1.n==0) return s2;
+  if(s2.n==0) return s1;
+  join_array(s1.n-s1.nzero, s1.part, s2.n-s2.nzero, s2.part, &s.n, &s.part);
+  join_array(s1.nzero, s1.zeropart, s2.nzero, s2.zeropart, &s.nzero, &s.zeropart);
+  s.n=s.n+s.nzero;
+  if(s.n-s.nzero>0 && LAST(s)-s.part + 1 != s.n-s.nzero) ENDRUN("join error 1");
+  if(s.nzero>0 && LASTZERO(s)-s.zeropart + 1 != s.nzero) ENDRUN("join error 2");
   return s;
 }
 
 FLOAT global_timestep(struct sys s)
 {
-  UINT i;
-  FLOAT mindt;
-  mindt=HUGE_VAL;
-  for(i=0;i<s.n;i++)
+  FLOAT dt,mindt=HUGE_VAL;
+  for(UINT i=0;i<s.n;i++)
   {
-    if(mindt>s.part[i].timestep) mindt=s.part[i].timestep;
+    dt=GETPART(s, i)->timestep;
+    if(dt<mindt) mindt=dt;
   }
   return mindt;
 }
 
+FLOAT max_global_timestep(struct sys s)
+{
+  FLOAT dt,maxdt=0;
+  for(UINT i=0;i<s.n;i++)
+  {
+    dt=GETPART(s, i)->timestep;
+    if(dt>maxdt) maxdt=dt;
+  }
+  return maxdt;
+}
+
 void kdk(int clevel,struct sys s1,struct sys s2, DOUBLE stime, DOUBLE etime, DOUBLE dt)
 {
   if(s2.n>0) kick(clevel,s2, s1, dt/2);
@@ -647,3 +718,52 @@ void dkd(int clevel,struct sys s1,struct sys s2, DOUBLE stime, DOUBLE etime, DOU
   if(s2.n>0) kick(clevel,s2, s1, dt);
   drift(clevel,s1,etime, dt/2);
 }
+
+void split_zeromass(struct sys *s)
+{
+  UINT i=0;
+  struct particle *left, *right;
+  if(s->n==0) return;
+  if(s->part==NULL) ENDRUN("split_zeromass malformed input");
+  if(s->n-s->nzero==0)
+  {
+    if(s->zeropart==NULL || s->part!=s->zeropart) ENDRUN("split_zeromass malformed input");
+    if(LASTZERO(*s)-s->zeropart+1!=s->nzero) ENDRUN( "split_zeromass malformed input sys");
+    return;
+  }  
+  if(s->nzero!=0 && LAST(*s)+1!=s->zeropart) 
+    ENDRUN("split_zeromass can only work on fully contiguous sys");
+  left=s->part;
+  right=s->part+(s->n-1);
+  while(1)
+  {
+    if(i>=s->n) ENDRUN("split_zeromass error 1");
+    i++;
+    while(left->mass!=0 && left<right) left++;
+    while(right->mass==0 && left<right) right--;
+    if(left<right) 
+      {SWAP( *left, *right, struct particle);}
+    else 
+      break;  
+  }
+  if(left->mass!=0) left++;
+  s->nzero=s->n-(left-s->part);
+  if(s->nzero<0) ENDRUN("split_zeromass find negative number of part");
+  if(s->nzero>0)
+  {
+    s->zeropart=left;
+  }
+  if((left-s->part)+s->nzero !=s->n) ENDRUN( "split_zeromass error 2");
+  for(i=0;i<(s->n-s->nzero);i++) if(GETPART(*s,i)->mass==0) ENDRUN ("split_zromass error 3");
+  for(i=s->n-s->nzero;i<s->n;i++) if(GETPART(*s,i)->mass!=0) ENDRUN ("split_zeromass error 4");
+#ifdef CONSISTENCY_CHECKS
+  verify_split_zeromass(*s);
+#endif
+}
+
+void verify_split_zeromass(struct sys s)
+{
+  if(!accel_zero_mass) return;
+  for(UINT i=0;i<s.n-s.nzero;i++) if(GETPART(s,i)->mass==0) ENDRUN("massless particle in main part\n") 
+  for(UINT i=s.n-s.nzero;i<s.n;i++) if(GETPART(s,i)->mass!=0) ENDRUN("massive particle in massless part\n") 
+}
